/***************************************************
*                                                  *
*  Constraints File :                              *
*                                                  *
*  Contains all the constraints wanted on the      *
*  program.                                        *
*                                                  *
***************************************************/

/* --------------- User properties --------------- *
*                                                  *
* The user can define some properties. When they   *
* are called in an expression, the system check if *
* the property is true at the current node.        *
*                                                  *
* - Must have one parameter : the current node     *
*                                                  *
***************************************************/

% True if Here is a class
  is_class(Here) :- scope(Here,classtype(_,_)).
  is_class(Here) :- scope(Here,interfacetype(_,_)).

% True if Here is an attribute
  is_attribute(Here) :- scope(Here,field(_,_)).

% True if Here is a method
  is_method(Here) :- scope(Here,tuple(_,_)).

% True if Here is a member
  is_member(Here) :- is_attribute(Here).
  is_member(Here) :- is_method(Here).

/* ----- User logical expressions predicates ----- *
*                                                  *
* The user can define its own predicates to        *
* simplify the reading of its constraints.         *
*                                                  *
* - Must have two parameters (nodes)               *
* - Must call the predicate 'evaluate_expression'  *
*   with :                                         *
*     - the logical expression as first parameter, *
*     - 'true' as second parameter to tell that we *
*       are looking for models which check the     *
*       expression                                 *
*     - _ as last parameter beacause the final     *
*       model answer does not include the user     *
*       predicate model (a possibility to add      *
*       would be to ask with a parameter if the    *
*       user wants the full model (based on the    *
*       basic predicates 'u', 'c', 's') or the     *
*       partial model (based on all the defined    *
*       predicates, this is the one used here)     *
*                                                  *
* Ex : use(Node1,Node2) :- evaluate_expression(    *
*                           [arobase(Node1),       *
*                            diamond(u,[]),        *
*                            hereIs(Node2)],       *
*                          'true',                 *
*                          _).                     *
*                                                  *
***************************************************/

% Defines a complex use relation
complex_use(Node1,Node2) :-
  evaluate_expression(
    [arobase(Node1),
     diamond(c,[*,+]),
     diamond(u,[]),
     diamond(c,[*,-,+]),
     hereIs(Node2)],
    'true',
    _).

% Define a way to get the container
dec_class(Node1,Node2) :-
  evaluate_expression(
    [arobase(Node1),
     diamond('c',['-']),
     hereIs(Node2)],
    'true',
    _).

/* -------------- User constraints --------------- *
*                                                  *
* The user can now define its constraints.         *
*                                                  *
* - All constraints predicates must be named       *
*   'constraint'                                   *
* - Must have one parameter : the logical          *
*   expression expressing a constraint             *
* - No body                                        *
* - A property call must be "p(property_name)"     *
*   without giving the parameter                   *
* - A user expression predicate must be use as a   *
*   basic predicate, giving the name without any   *
*   parameter.                                     *
*                                                  *
* Basic predicates : 'u' : use relation            *
*                    'c' : contain relation        *
*                    's' : subclass relation       *
*                                                  *
* Relations quantifiers : '*' : reflexivity        *
*                         '+' : transitivity       *
*                         '-' : reversibility      *
*                                                  *
* Set here position  : arobase(here)               *
*                                                  *
* Assert here position : hereIs(here)              *
*                                                  *
***************************************************/

/**************************************************/
/* User constraints tools                         */
/**************************************************/

getnames(L, Names) :- maplist(node_name, L, Names).
node_name(P, Name) :- scope(node(Name), P).

/* Special constraints */

constraint(S):-
  sconst(F1,F2),
  append(F1,[box('u',['-'])],Part),
  append(Part,F2,S).

/* Virtual Package */

contains(node(VP),node(C)) :-
  virtual_package(VP,CList),
  member(C,CList).

sconst([arobase(node(VP)),
        box('c',['+'])],
       [diamond('c',['-', '+']),
        hereIs(node(VP))]) :-
  virtual_package(VP,_).

/* Private Class */

virtual_package(VP,[C]) :- private_class(VP,C).

/* Private Subclasses */

virtual_package(VP,[C]):-
  private_subclasses(Class),
  inherits(node(C),node(Class)),
  concat(Class,C,VP).

/* Hidden Subclasses */

virtual_package(VP,[C]):-
  hidden_subclasses(VP,Class),
  inherits(node(C),node(Class)).

virtual_package_n('','').

virtual_package(N, Names) :- virtual_package_n(N, L),
                             getnames(L, Names).

/* Private */

constraint(
  [ arobase(node(Name))
  , box('u',['-'])
  , diamond('c',['-', '+'])
  , hereIs(DecClass)
  ])
:- private(Type),
   node_name(Type, Name),
   dec_class(node(Name), DecClass).
