/***************************************************
*          Logical Expressions Evaluator           *
* ------------------------------------------------ *
*                                                  *
* Definition :                                     *
* ==========                                       *
*                                                  *
* Evaluate a request written in the grammar logic. *
*                                                  *
* Using :                                          *
* =====                                            *
*                                                  *
*  evaluate_expression(+Expression,                *
*                      ?Result,                    *
*                      ?Model)                     *
*                                                  *
*    Expression :                                  *
*      Logical expression to process               *
*                                                  *
*    Result :                                      *
*      'true' or 'false' evaluation result value   *
*                                                  *
*    Model :                                       *
*      Partial or total model of the expression    *
*                                                  *
***************************************************/

% Import the graph relations tools library

  :- consult('graph_relations_tools.pl').

% Evaluate a logical expression
% +Expression : List representing the expression
% ?Result     : Result value 'true' or 'false'
% ?Model      : Partial or total expression model

  evaluate_expression(Expression,Result,Model) :-
	evaluate('',Expression,Result,Model).

% Evaluate a logical expression from Here
% ?Here       : Node where to start from
% +Expression : Expression to evaluate
% ?Result     : Result value
% ?Model      : Partial or total expression model

  % obsolete
  evaluate(Here,[not(Expr)],Result,Model) :-
	evaluate(Here,Expr,NotResult,Model),
	not_result(NotResult,Result).

  evaluate(Here,[not | Expr],Result,Model) :-
	evaluate(Here,Expr,NotResult,Model),
	not_result(NotResult,Result).
  evaluate(Here,[pred(Predicate, Expr)| NextExpr],Result,Model) :-
        call(Predicate, Expr, PredExpr),
        append(PredExpr, NextExpr, ResExpr),
	evaluate(Here,ResExpr,Result,Model).
  evaluate(_,[arobase(NewHere)|Expr],Result,[NewHere|Model]) :-
	evaluate(NewHere,Expr,Result,Model).
  evaluate(Here,[diamond(Relation,Params)|Expr],
           Result,[[Relation|Params]|Model]) :-
	evaluate_binary_relation(Here,Relation,Params,HereAndPathList),
	evaluate_diamond(HereAndPathList,Expr,Result,Model).
  evaluate(Here,[box(Relation,Params)|Expr],
           Result,[[Relation|Params]|Model]) :-
	evaluate_binary_relation(Here,Relation,Params,HereAndPathList),
	evaluate_box(HereAndPathList,Expr,Result,Model).
  evaluate(Here,[p(Property)|Expr],'true',[Property|Model]) :-
	call(Property,Here),
	evaluate(Here,Expr,'true',Model).
  evaluate(Here,[p(Property)|_],'false',['WRONG PROPERTY']) :-
	not(call(Property,Here)).
  evaluate(Here,[hereIs(Here)|Expr],Result,Model) :-
	evaluate(Here,Expr,Result,Model).
  evaluate(Here,[hereIs(There)|_],'false',['WRONG LOCATION']) :-
	Here \== There.
  evaluate(_,[],'true',[]).

% Evaluate a diamond
% +HereAndPathList     : Node where to start from
%                        and path to reach it
% +Expression          : Expression to evaluate
% ?Result              : Result value
% ?PathAndNewHereModel : Path to the new usable here

  evaluate_diamond([(NewHere,_)|HereAndPathList],Expr,'true',[PathAndNewHereAndModel]) :-
	append([NewHere],Model,PathAndNewHereAndModel),
	evaluate(NewHere,Expr,'true',Model),
	evaluate_diamond(HereAndPathList,Expr,'false',_).
  evaluate_diamond([(NewHere,_)|HereAndPathList],Expr,'true',Models) :-
	evaluate(NewHere,Expr,'false',_),
	evaluate_diamond(HereAndPathList,Expr,'true',Models).
  evaluate_diamond([(NewHere,_)|HereAndPathList],Expr,'true',[PathAndNewHereAndModel|Models]) :-
	append([NewHere],Model,PathAndNewHereAndModel),
	evaluate(NewHere,Expr,'true',Model),
	evaluate_diamond(HereAndPathList,Expr,'true',Models).
  evaluate_diamond([(NewHere,_)|HereAndPathList],Expr,'false',[PathAndNewHereAndModel|Models]) :-
	append([NewHere],Model,PathAndNewHereAndModel),
	evaluate(NewHere,Expr,'false',Model),
	evaluate_diamond(HereAndPathList,Expr,'false',Models).
  evaluate_diamond([],_,'false',[['EMPTY DIAMOND']]).

% Evaluate a box
% +HereAndPathList      : Node where to start from
%                         and path to reach it
% +Expression           : Expression to evaluate
% ?Result               : Result value
% ?PathAndHereAndModels : Paths to all the new usable here

  evaluate_box([(Here,_)|HereAndPathList],Expr,'false',[PathAndHereAndModel]) :-
	append([Here],Model,PathAndHereAndModel),
	evaluate(Here,Expr,'false',Model),
	evaluate_box(HereAndPathList,Expr,'true',_).
  evaluate_box([(Here,_)|HereAndPathList],Expr,'false',Models) :-
	evaluate(Here,Expr,'true',_),
	evaluate_box(HereAndPathList,Expr,'false',Models).
  evaluate_box([(Here,_)|HereAndPathList],Expr,'false',[PathAndHereAndModel|Models]) :-
	append([Here],Model,PathAndHereAndModel),
	evaluate(Here,Expr,'false',Model),
	evaluate_box(HereAndPathList,Expr,'false',Models).
  evaluate_box([(Here,_)|HereAndPathList],Expr,'true',[PathAndHereAndModel|Models]) :-
	append([Here],Model,PathAndHereAndModel),
	evaluate(Here,Expr,'true',Model),
	evaluate_box(HereAndPathList,Expr,'true',Models).
  evaluate_box([],_,'true',[]).

% Invert a boolean value
% ?Result1 : 'true' or 'false'
% ?Result2 : 'true' or 'false'

  not_result('true','false').
  not_result('false','true').

% Binary relation evaluation predicate interface
% +Here            : Node where to start from
% +Relation        : Relation to evaluate
% ?Params          : Params of the relation
% ?HereAndPathList : Node where to start from
%                    and path to reach it

  evaluate_binary_relation(Here,Relation,Params,HereAndPathList) :-
	evaluate_binary_relation(Here,Relation,Params,'','',HereAndPathList).

% Binary relation evaluation predicate interface
% +Here            : Node where to start from
% +Relation        : Relation to evaluate
% ?Params          : Params of the relation
% +Direction       : '' normal or '-' reverse
% +Transitivity    : '' normal or '+' transitivity
% ?HereAndPathList : Reachable nodes and path
%                    to reach them

% A - Evaluation without transitivity

  evaluate_binary_relation(Here,Relation,[],Dir,'',HereAndPathList) :-
	evaluate_predicate(Relation,Dir,Here,HereList),
	fill_here_and_path_list(HereList,[],HereAndPathList).

% B - Evaluation with transitivity

  evaluate_binary_relation(Here,Relation,[],Dir,'+',HereAndPathList) :-
	evaluate_predicate(Relation,Dir,Here,HereList),
	fill_transitivity_here_list(HereList,[],Relation,Dir,HereAndPathList1),
	fill_here_and_path_list(HereList,[],HereAndPathList0),
	append(HereAndPathList0,HereAndPathList1,HereAndPathList).

% Dagger : reverse relation

  evaluate_binary_relation(Here,Relation,['-'|Params],_,Trans,HereAndPathList) :-
	evaluate_binary_relation(Here,Relation,Params,'-',Trans,HereAndPathList).

% Star   : set reflexivity property

  evaluate_binary_relation(Here,Relation,['*'|Params],Dir,Trans,
                           [(Here,[])|HereAndPathList]) :-
	evaluate_binary_relation(Here,Relation,Params,Dir,Trans,HereAndPathList).

% Cross  : set transitivity property

  evaluate_binary_relation(Here,Relation,['+'|Params],Dir,_,HereAndPathList) :-
	evaluate_binary_relation(Here,Relation,Params,Dir,'+',HereAndPathList).

% Fill the list of possible here with the
% transitivity property for a given relation
% +HereListInit    : Here nodes where to start from
% +PathList        : Path to reach here
% +Relation        : Relation to evaluate
% +Dir             : '' normal or '-' reverse
% ?HereAndPathList : Reachable nodes and path
%                    to reach them

  fill_transitivity_here_list([Here|HereListInit],PathList,
                              Relation,Dir,HereAndPathList) :-
	evaluate_predicate(Relation,Dir,Here,HereList0),
	append(PathList,[[Here]],PathListHere),
	fill_here_and_path_list(HereList0,PathListHere,HereAndPathList0),
	fill_transitivity_here_list(HereListInit,PathList,
                                    Relation,Dir,HereAndPathList1),
	fill_transitivity_here_list(HereList0,PathListHere,
                                    Relation,Dir,HereAndPathList2),
	append(HereAndPathList0,HereAndPathList1,HereAndPathListTmp),
	append(HereAndPathListTmp,HereAndPathList2,HereAndPathList).
  fill_transitivity_here_list([],_,_,_,[]).

% Create for an here list and a path list
% the here and path list
% ?HereList        : Nodes
% ?PathList        : Path to these nodes
% ?HereAndPathList : Couple (Here,Path)

  fill_here_and_path_list([Here|HereList],[Path|PathList],
                          [(Here,Path)|HereAndPathList]) :-
	fill_here_and_path_list(HereList,PathList,HereAndPathList).
  fill_here_and_path_list([Here|HereList],[],
                          [(Here,[])|HereAndPathList]) :-
	fill_here_and_path_list(HereList,[],HereAndPathList).
  fill_here_and_path_list([],_,[]).

% Evaluate a predicate in normal or reverse direction
% +Relation  : Relation to evaluate
% +Direction : '' normal or '-' reverse
% +Here      : Node where to start from
% ?ThereList : List of reachable nodes

  evaluate_predicate(Relation,'',Here,ThereList) :-
	findall(There,call(Relation,Here,There),ThereList).
  evaluate_predicate(Relation,'-',Here,ThereList) :-
	findall(There,call(Relation,There,Here),ThereList).

/**************************************************/
/* Extract violation uses arcs                    */
/**************************************************/

/* Extract violation arcs */

  extract_arcs([],[]).
  extract_arcs([(Const,Viols)|Reste],
	       [violation(Const,CArcs)|ResteArcs]) :-
    extract_path(Viols,Paths),
    extract_uses(Paths, Arcs),
    extract_arcs(Reste,ResteArcs),
    clean_list(Arcs,CArcs).

/* Extract violation arcs paths */

  extract_path([],[[]]).
  extract_path([X,'WRONG LOCATION'],[[X,'WRONG LOCATION']]).
  extract_path([X,'WRONG PROPERTY'],[[X,'WRONG PROPERTY']]).
  extract_path(['EMPTY DIAMOND'],[['EMPTY DIAMOND']]).

  extract_path([_, _], []).

  extract_path([X, R, Y| Tail], Out) :-
    extract_path([X, R| Tail], Out1),
    extract_path(Y, Out2),
    extract_tpath([X, R], Out2, Out3),
    append(Out1, Out3, Out).

  extract_tpath([_,_],[],[]).
  extract_tpath([_,_],[[]],[]).
  extract_tpath([X,R],[P],[[X,R|P]]).
  extract_tpath([X,R],[P|Reste],[[X,R|P]|RPath]):-
   extract_tpath([X,R],Reste,RPath).

/* Extract violation uses arcs */

  extract_uses([],[[]]).
  extract_uses([L|Reste],[LResult|RResult]):-
    extract_use(L,LResult),
    extract_uses(Reste,RResult).

/* Extract violation uses arc */

  extract_use([],[]).
  extract_use([_],[]).
  extract_use([_,_],[]).
  extract_use([X, [u, '-'], Y| Reste],
	      [uses(Y,X)|Arcs]):-
    !,
    extract_use([Y|Reste],Arcs).
  extract_use([X,[u,[]],Y| Reste],
              [uses(X,Y)|Arcs]):-
    !,
    extract_use([Y|Reste],Arcs).
  extract_use([_,[u|_],Y| Reste],
	      Arcs):-
    !,
    extract_use([Y|Reste],Arcs).
  extract_use([_,[_|_],Y| Reste],
	      Arcs):-
    extract_use([Y|Reste],Arcs).

/* Remove similar arcs from list */

clean_list([],[]).
clean_list([N|R],CR):-
  member(N,R),
  clean_list(R,CR).
clean_list([N|R],[N|CR]):-
  not(member(N,R)),
  clean_list(R,CR).
