/***************************************************
*     Decoupling Constraints Program Checker       *
* ------------------------------------------------ *
*                                                  *
* Definition :                                     *
* ==========                                       *
*                                                  *
* Check if the given program relations are correct *
* according to the given decoupling constraints.   *
*                                                  *
* Using :                                          *
* =====                                            *
*                                                  *
*  check_program(+RelationsFile,                   *
*                +ConstraintsFile,                 *
*                -Violations)                      *
*                                                  *
*    RelationsFile :                               *
*      File containing all the graph relations     *
*                                                  *
*    ConstraintsFile :                             *
*      File containing all the constraints         *
*                                                  *
*    Violations :                                  *
*      Violated contraints list                    *
*                                                  *
***************************************************/

% Declare discontiguous basic relations

  :- discontiguous(scope/2).
  :- discontiguous(node/1).

  :- discontiguous(classtype/2).
  :- discontiguous(interfacetype/2).
  :- discontiguous(typename/1).
  :- discontiguous(typename/2).
  :- discontiguous(id/1).
  :- discontiguous(tuple/2).
  :- discontiguous(method/1).
  :- discontiguous(method/3).
  :- discontiguous(methodname/2).
  :- discontiguous(methodname/1).
  :- discontiguous(packagename/1).
  :- discontiguous(field/2).
  :- discontiguous(exprname/1).
  :- discontiguous(exprname/2).
  :- discontiguous(some/1).
  :- discontiguous(none/0).
  :- discontiguous(root/0).
  :- discontiguous(arraytype/1).
  :- discontiguous(arrayvardecid/2).
  :- discontiguous(dim/0).

  :- discontiguous(inherits/2).
  :- discontiguous(contains/2).
  :- discontiguous(uses/2).

  :- discontiguous(constraint/1).
  :- discontiguous(virtual_package/2).
  :- discontiguous(virtual_package_n/2).
  :- discontiguous(private_class/2).
  :- discontiguous(private_subclasses/1).
  :- discontiguous(hidden_subclasses/2).

  :- discontiguous(private/1).
  :- discontiguous(protected/1).
  :- discontiguous(public/1).
  :- discontiguous(final/1).

% Declare multifile basic relations

  :- multifile(scope/2).
  :- multifile(node/1).

  :- multifile(classtype/2).
  :- multifile(interfacetype/2).
  :- multifile(typename/1).
  :- multifile(typename/2).
  :- multifile(id/1).
  :- multifile(tuple/2).
  :- multifile(method/1).
  :- multifile(method/3).
  :- multifile(methodname/2).
  :- multifile(methodname/1).
  :- multifile(packagename/1).
  :- multifile(field/2).
  :- multifile(exprname/1).
  :- multifile(exprname/2).
  :- multifile(some/1).
  :- multifile(none/0).
  :- multifile(root/0).
  :- multifile(arraytype/1).
  :- multifile(arrayvardecid/2).
  :- multifile(dim/0).

  :- multifile(inherits/2).
  :- multifile(contains/2).
  :- multifile(uses/2).

  :- multifile(constraint/1).
  :- multifile(virtual_package/2).
  :- multifile(virtual_package_n/2).
  :- multifile(private_class/2).
  :- multifile(private_subclasses/1).
  :- multifile(hidden_subclasses/2).

  :- multifile(private/1).
  :- multifile(protected/1).
  :- multifile(public/1).
  :- multifile(final/1).

% Assert that the prediacte 'constraint' exists

  constraint('').

% Evaluator of the logic expressions
% used to express constraints

  :- consult('le_evaluator.pl').

% Check a program, given by its graph relations,
% according to the given constraints
% +RelationsFile   : File containing the relations
%                    of the graph of the program
% +ConstraintsFile : File containing the constraints
%                    to check
% ?Violations      : Couples (Violated constraint, Model)

  check_program(RelationsFile,ConstraintsFile,Violations,Out) :-

    % Load the program graph relations
    % Load the constraints to respect
    load_files([RelationsFile, ConstraintsFile],[silent(true)]),

    % Get all the constraints
    findall(Constraint,constraint(Constraint),Constraints),

    % Check them
    findall((Const,Violation),
	       (member(Const,Constraints),
                evaluate_expression(Const,'false',Violation)),
	     Violations),
    extract_arcs(Violations,Results),
    write_term(Out,Results,[quoted(true)]),
    write_term(Out,'.',[quoted(false)]).
