/***************************************************************************
* Java access graph constraints
* 
* Format and manage constraints for violations correction
*
***************************************************************************/

%--------------------------------------------------------------------------
% Predicates declarations
%--------------------------------------------------------------------------
:- discontiguous(hidden/1).
:- discontiguous(hidden/2).
:- discontiguous(hiddenClass/2).
:- discontiguous(hiddenSubclass/2).
:- discontiguous(virtualPackage/1).
:- discontiguous(virtual_package_n/2).
:- discontiguous(virtual_package/2).
:- discontiguous(private/1).
:- discontiguous(virtualPackage/1).

:- discontiguous(forbidUsesConstr/2).
:- discontiguous(authorUsesConstr/2).

:- discontiguous(classSet/1).

:- dynamic(forbidUsesConstr/2).
:- dynamic(authorUsesConstr/2).
:- dynamic(usesConstraint/3).
:- dynamic(classSet/1).

%--------------------------------------------------------------------------
% Graph constraints definition
%
% graphConstraints(-converted_graph_constr_list)
%--------------------------------------------------------------------------
graphConstraints(ConvGraphConstrLst) :-
  ConstraintsList = [                                     % Define graph constraint
  [inherits,class,class],
  [inherits,class,interface],
  [contains,package,class],
  [contains,package,interface],
  [contains,class,class],
  [contains,class,constructor],
  [contains,class,method],
  [contains,class,field],
  [contains,interface,method],
  [contains,method,class],
  [uses,constructor,class],
  [uses,constructor,constructor],
  [uses,constructor,method],
  [uses,constructor,field],
  [uses,constructor,type],
  [uses,method,class],
  [uses,method,interface],
  [uses,method,constructor],
  [uses,method,method],
  [uses,method,field],
  [uses,method,type],
  [uses,field,class],
  [uses,field,interface],
  [uses,field,constructor],
  [uses,field,method],
  [uses,field,type]
  ],                                                      
  graphConstrListConversion(ConstraintsList,ConvGraphConstrLst). % Convert graph constraints list

%--------------------------------------------------------------------------
% Decoupling constraints definition
%
% forbidUsesConstr(-SrcElement,-DestElement)
% authorUsesConstr(-SrcElement,-DestElement)
%--------------------------------------------------------------------------

% hidden/1
forbidUsesConstr(all,Element) :-
  hidden(Element).
authorUsesConstr(Element,Element) :-
  hidden(Element).

% hidden/2
forbidUsesConstr(Element1,Element2) :-
  hidden(Element1,Element2).

% hiddenClass/2
forbidUsesConstr(Element1,ElementLst) :-
  hiddenClass(Element1,ElementLst).

% virtualPackage/1
forbidUsesConstr(all,ElementLst) :-
  virtualPackage(ElementLst).
authorUsesConstr(ElementLst,ElementLst) :-
  virtualPackage(ElementLst).
classSet(ElementLst) :-
  virtualPackage(ElementLst).

% virtual_package_n/2
forbidUsesConstr(all,ElementLst) :-
  virtual_package_n(_,ElementLst).
authorUsesConstr(ElementLst,ElementLst) :-
  virtual_package_n(_,ElementLst).
classSet(ElementLst) :-
  virtual_package_n(_,ElementLst).

% virtual_package/2
forbidUsesConstr(all,ElementLst) :-
  virtual_package(_,ElementLst).
authorUsesConstr(ElementLst,ElementLst) :-
  virtual_package(_,ElementLst).
classSet(ElementLst) :-
  virtual_package(_,ElementLst).

% hiddenSubclass/2
forbidUsesConstr(Element1,subClass(Element2)) :-
  hiddenSubclass(Element1,Element2).

% private/1
forbidUsesConstr(all,Element) :-
  private(Element).
authorUsesConstr(class(Element),Element) :-
  private(Element).

%--------------------------------------------------------------------------
% Constraints initialisation
%
% constraintsInit()
% constrLstProcess(+Constraints_list,+Constraint_type)
%                 Constraint_type in (forbid,authorize)
% testSubClass(+Element,-SubClass_of_element_list)
% testClass(+Element,-Class_contains_elements_list)
%--------------------------------------------------------------------------
constraintsInit :-
  retractall(usesConstraint(_,_,_)),                      % Remove previous defined usesConstraints
  findall((Elem1,Elem2),forbidUsesConstr(Elem1,Elem2),ForbidConstrList),  % Get all forbidUsesConstr
  constrLstProcess(ForbidConstrList,forbid),              % Create usesConstraints
  findall((Elem1,Elem2),authorUsesConstr(Elem1,Elem2),AuthorConstrList),  % Get all authorUsesConstr
  constrLstProcess(AuthorConstrList,authorize).           % Create usesConstraints

constrLstProcess([],_).
constrLstProcess([(Elem1,Elem2)|NextElemLst],Type) :-
  (testClass(Elem1,NewElem1);true),                       % If Elem1 is class(Elem1)
  (testClass(Elem2,NewElem2);true),                       % If Elem2 is class(Elem2)
  (testSubClass(Elem1,NewElem1);true),                    % If Elem1 is subClasse(Elem1)
  (testSubClass(Elem2,NewElem2);true),                    % If Elem2 is subClasse(Elem2)
  !,
  (var(NewElem1) *->                                      % If NewElem1 is free then
    (is_list(Elem1),var(NewElem1) *->                     % If Elem1 is list then
      NewElem1=Elem1                                      
      ;NewElem1=[Elem1]                                   % Formet Elem1 to list
    )
    ;true                                                 % Else true
  ),
  (var(NewElem2) *->                                      % If NewElem2 is free then
    (is_list(Elem2),var(NewElem2) *->                     % If Elem2 is list
      NewElem2=Elem2
      ;NewElem2=[Elem2]                                   % Format Elem2 to list
    )
    ;true
  ),
  lstCartesProduct(NewElem1,NewElem2,Type),               % Cartesian product of two list
  constrLstProcess(NextElemLst,Type).                     % Next elements

testSubClass(subClass(Element),subClassLst) :-        
  scope(ElementId,Element),
  findall(InheritsElement,
    (inherits(InheritsElementId,ElementId),scope(InheritsElementId,InheritsElement))
    ,subClassLst).

testClass(class(Element),Class) :-
  scope(ElementId,Element),
  findall(ClassElement,
    (contains(ClassElementId,ElementId),scope(ClassElementId,ClassElement))
    ,Class).

%--------------------------------------------------------------------------
% Lists cartesian product
%
% lstCartesProduct(+Element_list,+Element_list,+Constraint_type)
% subLstCartesProduct(+Element,+Element_list,+Constraint_type)
%--------------------------------------------------------------------------
lstCartesProduct([],_,_).
lstCartesProduct([FirstElemLst|NextElemLst],ElemLst,Type) :-
  subLstCartesProduct(FirstElemLst,ElemLst,Type),
  lstCartesProduct(NextElemLst,ElemLst,Type).

subLstCartesProduct(_,[],_).
subLstCartesProduct(Elem,[FirstElemLst|NextElemLst],Type) :-
  (Elem=all *->                                           % If Elem = all term
    ConvElem1=11                                          % Define integer conversion to 11
    ;scope(Elem1Name,Elem),transNode(ConvElem1,Elem1Name) % Else get Elem_id
  ),
  (FirstElemLst=all *->                                   % If Elem = all term
    ConvElem2=11                                          % Define integer conversion to 11
    ;scope(Elem2Name,FirstElemLst),transNode(ConvElem2,Elem2Name) % Else get Elem_id
  ),
  assert(usesConstraint(Type,ConvElem1,ConvElem2)),       % Add usesConstraints fact
  subLstCartesProduct(Elem,NextElemLst,Type).             % Next element

%--------------------------------------------------------------------------
% Test if there is a constraint for two elements
%
% isThereConstraints(+Constraint_type,+Element,+Element,-Constraint)
%--------------------------------------------------------------------------
isThereConstraints(Type,Element1,Element2,Constraint) :-
  usesConstraint(Type,ConstrElement1,ConstrElement2),
  (ConstrElement1=11 *->                                  % If all term 
    true                                                  % Then constraint exist
    ;pathExist(ConstrElement1,Element1)                   % Else test path between ConstrElem
  ),                                                      % and element
  (ConstrElement2=11 *->                                  % If all term
    true                                                  % Then constraint exist
    ;pathExist(ConstrElement2,Element2)                   % Else test path
  ),
  !,
  Constraint=[[ConstrElement1,ConstrElement2]].           % Return match constraint

isThereConstraints(_,_,_,[[]]).


%--------------------------------------------------------------------------
% Get and verify all possible graph relations between two elements
%
% getAllPossibleGraphRelation(+-Relation_type,+-Element1_type,+-Element2_type)
%--------------------------------------------------------------------------
getAllPossibleGraphRelation(RelationType,ElemType1,ElemType2) :-
  graphConstraints(GraphConstrLst),                       % Get graph constraints list
  GraphRelation = [[RelationType,ElemType1,ElemType2]],   % Define graph relation
  tuples_in(GraphRelation,GraphConstrLst).                % Test if relation is in graph
                                                          % constraints list


%--------------------------------------------------------------------------
% Verify if there is a constraint for a relation
%
% verifyConstrRelation(+Src_element,+Dst_element)
%--------------------------------------------------------------------------
verifyConstrRelation(SrcElem,DstElem) :-
  % Find all useConstraint
  findall([Elem1,Elem2],usesConstraint(forbid,Elem1,Elem2),ForbidConstrList),
  findall([Elem1,Elem2],usesConstraint(authorize,Elem1,Elem2),AuthorizeConstrList),
  % Test is constraint exist for relation
  isThereConstraints(forbid,SrcElem,DstElem,RelationForbidConstr),
  isThereConstraints(authorize,SrcElem,DstElem,RelationAuthorizeConstr),
  % Verify if relation constraint is in useConstraint
  (tuples_in(RelationAuthorizeConstr,AuthorizeConstrList)
  ;\+tuples_in(RelationForbidConstr,ForbidConstrList)).

