/***************************************************************************
* Java access graph conversion 
*
* Convert an access graph in integer for constraints solver use
*
***************************************************************************/

%--------------------------------------------------------------------------
% Predicates declarations 
%--------------------------------------------------------------------------
:- dynamic(convNode/2).
:- dynamic(convInherits/2).
:- dynamic(convContains/2).
:- dynamic(convUses/2).
:- dynamic(convClassSet/1).

:- dynamic(inherits/2).
:- dynamic(contains/2).
:- dynamic(uses/2).

%--------------------------------------------------------------------------
% Set conversion for elements types & relations
%--------------------------------------------------------------------------
typeConversion(1,package).
typeConversion(2,class).
typeConversion(3,interface).
typeConversion(4,constructor).
typeConversion(5,method).
typeConversion(6,field).
typeConversion(7,type).

relationConversion(1,inherits).
relationConversion(2,contains).
relationConversion(3,uses).

%--------------------------------------------------------------------------
% Acces graph conversion 
%
% graphConversion()
% nodeConversion()      : convert node fact
% inheritsConversion()  : convert inherits fact
% containsConversion()  : convert contains fact
% usesConversion()      : convert uses fact
% classSetConversion () : convert classSet fact
%--------------------------------------------------------------------------
graphConversion :-
  \+ nodeConversion,
  \+ inheritsConversion,
  \+ containsConversion,
  \+ usesConversion,
  (classSetConversion;true).

%--------------------------------------------------------------------------
% Input :   scope(Node_id, Node_type_name)
% Output :  convNode(Integer_node_id, Node_type)
%           transNode(Integer_node_id, Node_id)
%--------------------------------------------------------------------------
nodeConversion :-
  scope(NodeId,NodeTypeName),                             % For all scope
  getNodeType(NodeTypeName,Type),                         % Get node type
  gensym('99',StrId),                                     % Generate unique id beginning by 99
  ato-to-int(StrId,Id),                                   % Convert string id to integer
  assert(convNode(Id,Type)),                              % Add convNode fact 
  assert(transNode(Id,NodeId)),                           % Add transNode fact 
  fail.                                                   % Repeat for all scope

%--------------------------------------------------------------------------
% Input :   inherits(Node_id1,Node_id2)
% Output :  convInherits(Integer_node_id1,Integer_node_id2)
%--------------------------------------------------------------------------
inheritsConversion :-
  inherits(Elem1,Elem2),                                  % Get inherits relation
  transNode(Elem1Id,Elem1),                               % Get id of elements 
  transNode(Elem2Id,Elem2),
  assert(convInherits(Elem1Id,Elem2Id)),                  % Add convInherits fact 
  fail.                                                   % Repeat for all inherits relation

%--------------------------------------------------------------------------
% Input :   contains(Node_id1,Node_id2)
% Output :  convContains(Integer_node_id1,Integer_node_id2)
%--------------------------------------------------------------------------
containsConversion :-
  contains(Elem1,Elem2),                                  % Get containts relation
  transNode(Elem1Id,Elem1),                               % Get id of elements
  transNode(Elem2Id,Elem2),
  assert(convContains(Elem1Id,Elem2Id)),                  % Add convContains fact 
  fail.                                                   % Repeat for all contains relation

%--------------------------------------------------------------------------
% Input :   uses(Node_id1,Node_id2)
% Output :  convUses(Integer_node_id1,Integer_node_id2)
%--------------------------------------------------------------------------
usesConversion :-
  uses(Elem1,Elem2),                                      % Get uses relation
  transNode(Elem1Id,Elem1),                               % Get id of elements
  transNode(Elem2Id,Elem2), 
  assert(convUses(Elem1Id,Elem2Id)),                      % Add convUses fact 
  fail.                                                   % Repeat for all uses relations

%--------------------------------------------------------------------------
% Input :   uses(Node_id1,Node_id2)
% Output :  convUses(Integer_node_id1,Integer_node_id2)
%--------------------------------------------------------------------------
classSetConversion :-                                     
  classSet(ElemLst),                                      % Get all classSet fact 
  convClassSetLst(ElemLst,ConvElemLst),                   % Convert all elements of classSet
  assert(convClassSet(ConvElemLst)).                      % Add convClassSet fact

convClassSetLst([],[]).
convClassSetLst([FirstElem|NextElem],[FirstElemConver|NextElemConver]) :-
  nameToId(FirstElem,FirstElemConver),
  convClassSetLst(NextElem,NextElemConver).

%--------------------------------------------------------------------------
% Get Node_type_name of Integer_node_id 
% nameToId(+-Node_type_name,+-Integer_node_id)
%--------------------------------------------------------------------------
nameToId(NodeTypeName,IntNodeId) :-
  scope(NodeId,NodeTypeName),
  transNode(IntNodeId,NodeId).

%--------------------------------------------------------------------------
% Get Integer_node_id from Node_type_name or Node_Id
% NameOrNodeToId(+-Node_type_name,+-Integer_node_id)
% NameOrNodeToId(+-Node_Id,+-Integer_node_id)
%--------------------------------------------------------------------------
nameOrNodeToId(NodeTypeName,IntNodeId) :-
  nameToId(NodeTypeName,IntNodeId).

nameOrNodeToId(NodeId,IntNodeId) :-
  transNode(IntNodeId,NodeId).

%--------------------------------------------------------------------------
% Atom to int conversion
% 
% ato-to-int(+Atom,-Integer)
%--------------------------------------------------------------------------
ato-to-int(Ato,Int) :-
   term_to_atom(Int,Ato), integer(Int).

%--------------------------------------------------------------------------
% Get Node type
% getNodeType(Node_type_name,Integer_type)
%--------------------------------------------------------------------------
getNodeType(packagename(_),TypeConver):-                  % Package
  typeConversion(TypeConver,package).

getNodeType(classtype(_,_),TypeConver):-                  % Class
  typeConversion(TypeConver,class).

getNodeType(interfacetype(_,_),TypeConver):-              % Interface
  typeConversion(TypeConver,interface).

getNodeType(constructorname(_,_),TypeConver):-            % Constructor
  typeConversion(TypeConver,constructor).

getNodeType(methodname(_),TypeConver):-                   % MethodName
  typeConversion(TypeConver,method).
getNodeType(methodname(_,_),TypeConver):-
  typeConversion(TypeConver,method).
getNodeType(methodname(_,_,_,_),TypeConver):-
  typeConversion(TypeConver,method).
getNodeType(method(_),TypeConver):-
  typeConversion(TypeConver,method).
getNodeType(method(_,_,_),TypeConver):-
  typeConversion(TypeConver,method).


getNodeType(exprname(_),TypeConver):-                     % Expression / field
  typeConversion(TypeConver,field).
getNodeType(exprname(_,_),TypeConver):-
  typeConversion(TypeConver,field).
getNodeType(field(_,_),TypeConver):-
  typeConversion(TypeConver,field).

getNodeType(typename(_),TypeConver):-                     % Type
  typeConversion(TypeConver,type).
getNodeType(typename(_,_),TypeConver):-
  typeConversion(TypeConver,type).

%--------------------------------------------------------------------------
% Graph constraints list conversion 
%
% graphConstrListConversion(+Graph_constr_lst,-Converted_graph_constrl_lst)
% graphConstrConversion(+Graph_relation,-Converted_graph_relation)
%--------------------------------------------------------------------------
graphConstrListConversion([],[]).
graphConstrListConversion([FirstElem|NextElem],[FirstElemConver|NextElemConver]) :-
  graphConstrConversion(FirstElem,FirstElemConver), 
  graphConstrListConversion(NextElem,NextElemConver).

graphConstrConversion([Rel1,Elem1,Elem2],[ConverRel1,ConverElem1,ConverElem2]) :-
  relationConversion(ConverRel1,Rel1),
  typeConversion(ConverElem1,Elem1),
  typeConversion(ConverElem2,Elem2).


