/***************************************************************************
* Java access graph fold library 
***************************************************************************/

%--------------------------------------------------------------------------
% Predicates declarations
%--------------------------------------------------------------------------
:- dynamic(methodSet/3).


%--------------------------------------------------------------------------
% Method Set initialisation
% Test if methods can be set
%
% initMethodSet()
% followClassSet()
% getClassSetMethod(+Class_set,-Method_name_in_class_set)
%--------------------------------------------------------------------------
initMethodSet :-
  (followClassSet;true).                                  

followClassSet :-
  convClassSet(ClassSetId),                               % Get classSet
  findall(Method,getClassSetMethod(ClassSetId,Method),MethodLst), % Find all ClassSet methods
  list_to_set(MethodLst,MethodLstWithoutDupl),            % Remove duplicate methods
  createMethodSet(ClassSetId,MethodLstWithoutDupl).       % Create method Set

getClassSetMethod(ClassSetId,Method):-                    % Get method of given class Set
  convContains(Class,Method),
  (convNode(Method,5);convNode(Method,4)),
  member(Class,ClassSetId).  

%--------------------------------------------------------------------------
% Create methodSet with method
%
% createMethodSet(+Class_set,-Method_name)
%--------------------------------------------------------------------------
createMethodSet(_,[]).
createMethodSet(ClassSet,[FirstElem|NextElem]) :-
  transNode(FirstElem,FirstElemName),                     
  scope(FirstElemName,methodname(_,MethodName,_,_)),      % Get method name
  findall(MethodId,                                       % Find all method with same name
          (scope(Method,methodname(_,MethodName,_,_)),
          transNode(MethodId,Method)),MethodLst),
  length(MethodLst,MethodLstLength),                      % Get number of methods
  ((MethodLstLength>=2,\+methodSet(ClassSet,MethodLst,0)) *-> % If number >2
    assert(methodSet(ClassSet,MethodLst,0))               % Then create method Set
    ;true),
  createMethodSet(ClassSet,NextElem).                     % Next method of list

createMethodSet(ClassSet,[FirstElem|NextElem]) :-
  transNode(FirstElem,FirstElemName),                     
  scope(FirstElemName,constructorname(_,_)),              % Create methodset for constructor
  assert(methodSet(ClassSet,[FirstElem],0)),
  createMethodSet(ClassSet,NextElem).                     % Next method of list

%--------------------------------------------------------------------------
% Test if an element is in classSet
%
% inClassSet(+Intermed_class,+Dst_elem)
%--------------------------------------------------------------------------
inClassSet(LinkElem,DstElem) :-
  convClassSet(ClassSet),                                 % Get all classSet
  convInherits(ClassInter,LinkElem),                      % Get class of LinkElem
  member(ClassInter,ClassSet),                            % If previous class is in classSet
  pathExist(ClassInterDst,DstElem),                       % And Class of DstElem 
  member(ClassInterDst,ClassSet),                         % Is in classSet too
  (convInherits(ClassInterDst,LinkElem) *->               % If inheritance link not exist
    true
    ;assert(convInherits(ClassInterDst,LinkElem))).       % Then add convInherits relation 


%--------------------------------------------------------------------------
% Create fold
%
% createFold(+Src_element,+Dst_element)
%--------------------------------------------------------------------------
createFold(SrcElem,DstElem) :-
  setNewElemName,                                         % Set name for all new nodes
  utilises(SrcElem,DstElem,FoldElem),                     % Get elem used for fold
  getFoldType(SrcElem,DstElem,FoldType),                  % Get fold type
  getUsedElementLst(DstElem,FoldElem,UsedElem),           % Get used element
  getUsedRelationLst(SrcElem,DstElem,FoldElem,FoldType,UsedRel),   % Get used relation
  (completeMethodSet(SrcElem,DstElem,FoldType);true),     % Complete existing methodSet
  nameToId(SrcElemName,SrcElem),                          % Get element name
  nameToId(DstElemName,DstElem),                          % Get element name
  Relation=relation(use,SrcElemName,DstElemName),         % Creation Relation fact
  assert(fold(FoldType,Relation,UsedElem,UsedRel)).    % Add fold fact


%--------------------------------------------------------------------------
% Get fold type
%
% getFoldType(+Src_element,+Dst_element,-Fold_type)
%--------------------------------------------------------------------------
getFoldType(SrcElem,DstElem,FoldType) :-
  convUses(SrcElem,IntermedElem),                         % If two uses relation
  verifyConstrRelation(SrcElem,IntermedElem),             % satisfy constraints
  convUses(IntermedElem,DstElem),                         % then static fold
  verifyConstrRelation(IntermedElem,DstElem),             % satisfy constraints
  !,
  convNode(SrcElem,SrcElemType),                          % Get src elem type
  (
    ((SrcElemType=4;SrcElemType=5) ->                     % If src elem = method or constructor
      FoldType=static_code_fold)                          % then static code fold
    ;
    ((SrcElemType=2;SrcElemType=3) ->                     % If src elem= class or interface
      FoldType=static_data_fold)                          % then static
  ).

getFoldType(SrcElem,DstElem,FoldType) :-
  convNode(SrcElem,SrcElemType),                          % Else dynamic fold
  convNode(DstElem,DstElemType),
  (
    ((DstElemType=2;DstElemType=3) ->                     % If dst elem = class or interface
      FoldType=type)                                      % Then type fold
    ;
    ((SrcElemType=4;SrcElemType=5) ->                     % If src elem = method or constructor
      FoldType=dynamic_code_fold)                         % Then dynamic code fold
    ;
    ((SrcElemType=2;SrcElemType=3) ->                     % If src elem= class or interface
      FoldType=dynamic_data_fold)                         % Then dynamic data fold
  ).

%--------------------------------------------------------------------------
% Find if method is the interface method of a method set 
%
% completeMethodSet(+Src_element,+Dst_element,+Fold_type)
%--------------------------------------------------------------------------
completeMethodSet(SrcElem,DstElem,FoldType):-
  (FoldType=dynamic_code_fold;FoldType=dynamic_data_fold  % If dynamic fold
  ; FoldType=type),
  utilises(SrcElem,DstElem,IntermedElem),                 % Get fold element
  convContains(ClassIntermedElem,IntermedElem),           % Get fold element class
  convContains(ClassDstElem,DstElem),                     % Get dest element class
  methodSet(ClassSet,MethodSet,MethodId),                 % Get method set
  MethodId=0,                                             % If method set is "free"
  member(ClassDstElem,ClassSet),                          % Test if dst elem class is in classSet
  member(DstElem,MethodSet),                              % Test if fold elem class is in classSet
  convInherits(ClassDstElem,ClassIntermedElem),           % If dst elem class inherits fold elem class
  retractall(methodSet(ClassSet,MethodSet,MethodId)),     % Set method for method set
  assert(methodSet(ClassSet,MethodSet,IntermedElem)).

%--------------------------------------------------------------------------
% Set name for all created element
%
% setNewElemName()
%--------------------------------------------------------------------------
setNewElemName :-
  convNode(Id,TypeConver),
  \+ transNode(Id,_),                                     % Get all element without name
  typeConversion(TypeConver,Type),                        % Get type of new element
  gensym(Type,NewName),                                   % Generate unique id from type
  assert(transNode(Id,id(NewName))).                      % Add new transNode fact

%--------------------------------------------------------------------------
% Get used element list for given fold
%
% getUsedElementLst(+Dst_element,+Fold_element,-Used_elements_list) 
% convUsedElemLst(+Fold_elements_list,-Used_elements_list)
%--------------------------------------------------------------------------
getUsedElementLst(DstElem,FoldElem,UsedElem) :-
  path(IntermedElem,FoldElem,Path1),                      % Get all elements of fold
  path(IntermedElem,DstElem,Path2),
  union(Path1,Path2,UnionPath),
  convUsedElemLst(UnionPath,UsedElem).                    % Get name of elements id

convUsedElemLst([],[]).
convUsedElemLst([FirstElem|NextElem],[FirstElemConver|NextElemConver]) :-
  \+ nameToId(_,FirstElem),                               % Only for created elements 
  !,
  transNode(FirstElem,FirstElemName),                     % Get element name
  convNode(FirstElem,FirstElemType),                      % Get element type
  typeConversion(FirstElemType,FirstElemTypeName),        % get type name
  FirstElemConver=element(FirstElemTypeName,FirstElemName), % Set relation
  convUsedElemLst(NextElem,NextElemConver).               % Next Elem

convUsedElemLst([_|NextElem],UsedElem) :-
  convUsedElemLst(NextElem,UsedElem).                     % If not created element

%--------------------------------------------------------------------------
% Get used relation list for given fold
%
% getUsedRelationLst(+Src_element,+Dst_element,+Fold_element,-Used_relations_list)
% usedRelLst(+Src_element,+Dst_element,+Fold_elements_list,-Used_relations_list)
% useUsedRelationLst(+Element,+Other_elementsl_list,-Relation)
% containUsedRelationLst(+Element,+Other_elementsl_list,-Relation)
% inheritUsedRelationLst(+Element,+Other_elementsl_list,-Relation)
%--------------------------------------------------------------------------
getUsedRelationLst(SrcElem,DstElem,FoldElem,FoldType,UsedRel) :-
  path(IntermedElem,FoldElem,Path1),                      % Get all elements of fold
  path(IntermedElem,DstElem,Path2),
  usedRelLst(SrcElem,DstElem,FoldElem,Path1,Path1RelLst), % Get used relation
  flatten(Path1RelLst,Path1FlattenRelLst),
  usedRelLst(SrcElem,DstElem,FoldElem,Path2,Path2RelLst), % Get used relation
  flatten(Path2RelLst,Path2FlattenRelLst),
  union(Path1FlattenRelLst,Path2FlattenRelLst,UsedRelLst),% Join and remove duplicate relations
  (FoldType=dynamic_code_fold *->                         % Get virtual use relation
    getVirtualUsedRel(FoldType,DstElem,FoldElem,VirtUsedRel),    % Get virtual use relation
    union(UsedRelLst,VirtUsedRel,UsedRel);
    UsedRel=UsedRelLst).

usedRelLst(_,_,_,[],[]).
usedRelLst(SrcElem,DstElem,FoldElem,[FirstElem|NextElem],[FirstRelLst|NextRelLst]) :-
  % Find all use relation for an element
  findall(UseRelLst,useUsedRelationLst(FirstElem,SrcElem,DstElem,NextElem,UseRelLst),UseUsedRel),
  % Find all contains relation for an element
  findall(ContRelLst,containUsedRelationLst(FirstElem,FoldElem,ContRelLst),ContainUsedRel),
  % Find all inherits relation for an element
  findall(InhRelLst,inheritUsedRelationLst(FirstElem,DstElem,NextElem,InhRelLst),InhUsedRel),
  union(UseUsedRel,ContainUsedRel,TmpUsedRel),            % Join previous lists
  union(TmpUsedRel,InhUsedRel,FirstRelLst),
  usedRelLst(SrcElem,DstElem,FoldElem,NextElem,NextRelLst). % Next element

useUsedRelationLst(Elem,SrcElem,DstElem,ElemLst,RelLst) :-
  \+ nameToId(_,Elem),
  transNode(Elem,ElemName),
  convUses(Elem,Elem2),                                   % Get uses relation
  (Elem2=SrcElem;Elem2=DstElem),
  getNodeName(Elem2,Elem2Name),                           % Get element name
  RelLst= relation(use,ElemName,Elem2Name),
  \+ member(Elem2,ElemLst).                               % Only if other relation elem
                                                          % is not in other fold elem list
useUsedRelationLst(Elem,SrcElem,DstElem,ElemLst,RelLst) :-
  \+ nameToId(_,Elem),
  transNode(Elem,ElemName),
  convUses(Elem2,Elem),                                   % Get uses relation
  (Elem2=SrcElem;Elem2=DstElem),
  getNodeName(Elem2,Elem2Name),                           % Get element name
  RelLst= relation(use,Elem2Name,ElemName),
  \+ member(Elem2,ElemLst).                               % Only if other relation elem
                                                          % is not in other fold elem list
  
containUsedRelationLst(Elem,FoldElem,RelLst) :-
  \+ nameToId(_,Elem),
  getNodeName(Elem,ElemName),
  convContains(Elem,Elem2),                               % Find contain relation
  getNodeName(Elem2,Elem2Name),                           % Get element name
  RelLst= relation(contain,ElemName,Elem2Name),
  pathExist(Elem2,FoldElem).                              % Only if other relation elem
containUsedRelationLst(Elem,FoldElem,RelLst) :-
  \+ nameToId(_,Elem),
  getNodeName(Elem,ElemName),
  convContains(Elem2,Elem),                               % Find contain relation
  getNodeName(Elem2,Elem2Name),                           % Get element name
  RelLst= relation(contain,Elem2Name,ElemName),
  pathExist(Elem2,FoldElem).                              % Only if other relation elem

  
inheritUsedRelationLst(Elem,DstElem,ElemLst,RelLst) :-
  \+ nameToId(_,Elem),
  getNodeName(Elem,ElemName),
  convInherits(Elem,Elem2),
  pathExist(Elem2,DstElem),                               % Only inherits relation for actual fold 
  getNodeName(Elem2,Elem2Name),                           % Get element name
  RelLst= relation(inherit,ElemName,Elem2Name),
  \+ member(Elem2,ElemLst).                               % Only if other relation elem
                                                          % is not in other fold elem list
inheritUsedRelationLst(Elem,DstElem,ElemLst,RelLst) :-
  \+ nameToId(_,Elem),
  getNodeName(Elem,ElemName),
  convInherits(Elem2,Elem),
  pathExist(Elem2,DstElem),                               % Only inherits relation for actual fold
  getNodeName(Elem2,Elem2Name),                           % Get element name
  RelLst= relation(inherit,Elem2Name,ElemName),
  \+ member(Elem2,ElemLst).                               % Only if other relation elem
                                                          % is not in other fold elem list


%--------------------------------------------------------------------------
% Get virtual use for dynamic code fold
%
% getUsedRelationLst(+Fold_type,+Dst_element,+Fold_element,-Virtual_use_relation)
%--------------------------------------------------------------------------
getVirtualUsedRel(FoldType,DstElem,FoldElem,VirtlUsedRel):-
  FoldType=dynamic_code_fold,                             % If dynamic code fold
  getNodeName(FoldElem,FoldElemName),
  getNodeName(DstElem,DstElemName),
  VirtlUsedRel= [relation(virtualuse,FoldElemName,DstElemName)].     % Set virtual use relation

%--------------------------------------------------------------------------
% Get node name
%
% getNodeName(+-Elem_id,+-Elem_id)
%--------------------------------------------------------------------------
getNodeName(Id,Name) :-
  (nameToId(Name,Id) *-> nameToId(Name,Id) ;transNode(Id,Name)).


