/***************************************************************************
* Violation correction module 
***************************************************************************/

%--------------------------------------------------------------------------
% Import SWI-Prolog library 
%--------------------------------------------------------------------------
:- use_module(library(bounds)).
:- use_module(library(clp_distinct)).
:- use_module(library('clp/bounds')).

%--------------------------------------------------------------------------
% Predicates declarations
%--------------------------------------------------------------------------
:- discontiguous(uses/2).
:- discontiguous(contains/2).
:- discontiguous(inherits/2).
:- discontiguous(hidden/1).
:- discontiguous(hidden/2).
:- discontiguous(hiddenClass/2).
:- discontiguous(hiddenSubclass/2).
:- discontiguous(virtualPackage/1).
:- discontiguous(virtual_package_n/2).
:- discontiguous(virtual_package/2).
:- discontiguous(private/1).
:- discontiguous(virtualPackage/1).
:- discontiguous(forbidUsesConstr/2).
:- discontiguous(authorUsesConstr/2).
:- discontiguous(classSet/1).

:- multifile(hidden/1).
:- multifile(hidden/2).
:- multifile(hiddenClass/2).
:- multifile(hiddenSubclass/2).
:- multifile(virtualPackage/1).
:- multifile(virtual_package_n/2).
:- multifile(virtual_package/2).
:- multifile(private/1).
:- multifile(virtualPackage/1).
:- multifile(forbidUsesConstr/2).
:- multifile(authorUsesConstr/2).
:- multifile(violation/2).

:- dynamic(fold/4).


%--------------------------------------------------------------------------
% Import 
%--------------------------------------------------------------------------
:- consult('access_graph_conv.pl').
:- consult('access_graph_constraints.pl').
:- consult('access_graph_lib.pl').
:- consult('access_graph_fold.pl').

%--------------------------------------------------------------------------
% Manual test
%--------------------------------------------------------------------------
test :-
%  violations_correction('./tests/personne-graph.pl','./tests/personne-constraints.pl','./tests/personne-violations.pl',_).
  violations_correction('./tests/Editor-graph.pl','./tests/Editor-constraints.pl','./tests/Editor-violations.pl',_).


%--------------------------------------------------------------------------
% Violations correction
%
% violations_correction(+Relations_file,+Constraints_file,+Violations_list,-Out)
% getViolations(+Violations_list)
% processViolation(+Uses_violation)
%--------------------------------------------------------------------------
violations_correction(RelationsFile,ConstraintsFile,Violations,Out) :-
  load_files([RelationsFile, ConstraintsFile, Violations],[silent(true)]),
  graphConversion,                                        % Acces graph conversion
  constraintsInit,                                        % Constrains initialisation
  initMethodSet,                                          % Method Set initialisation
  ( violation(_,ViolationsLst)                            % Get detected violations
  | ViolationsLst = []),
  flatten(ViolationsLst,FlatViolationsLst),
  getViolations(FlatViolationsLst),                       % Correct all violations
  findall(fold(Type,UseRel,UsedElem,UsedRel)              % Get all folds
          ,fold(Type,UseRel,UsedElem,UsedRel)
          ,Results),
  write_term(Out,Results,[quoted(true)]),                 % write folds
  write_term(Out,'.',[quoted(false)]).                    % write folds


getViolations([]).
getViolations([FirstElemLst|NextElemLst]) :-
  processViolation(FirstElemLst),                         % Process each violation 
  getViolations(NextElemLst).                             % Next violations

processViolation(uses(ElemSrc,ElemDst)) :-
  nameOrNodeToId(ElemSrc,ElemSrcId),                      % Get src element id
  nameOrNodeToId(ElemDst,ElemDstId),                      % Get dst element id
  retractall(convUses(ElemSrcId,ElemDstId)),              % Remove uses relation
  correctViolation(ElemSrcId,ElemDstId).                  % Correct violation

%--------------------------------------------------------------------------
% Correct violation
%
% correctViolation(+Element_src,+Element_dst)
%--------------------------------------------------------------------------
correctViolation(SrcElem,DstElem) :-
  convNode(SrcElem,SrcElemType),                          % Get src element type
  convNode(DstElem,DstElemType),                          % Get dst element type
  gensym('99',StrNewElem),                                % Generate new id
  ato-to-int(StrNewElem,NewElem),                         
  % Insert fold
  insertFold(SrcElem,SrcElemType,DstElem,DstElemType,SrcElem,SrcElemType,NewElem),
  createFold(SrcElem,DstElem).                            % create Fold for program output

%--------------------------------------------------------------------------
% Insert fold 
%
% insertFold(+Src_element,+Src_element_type,+Dst_element,+Dst_element_type,
% +Intermediate_element,+Intermediate_element_type,+New_element)
%--------------------------------------------------------------------------
insertFold(SrcElem,SrcElemType,DstElem,DstElemType,IntermedElem,IntermedElemType,NewElem) :-
  % Try to insert uses
  (
    insertUses(SrcElem,DstElem,IntermedElem,IntermedElemType,NewElem,NewElemType)
    ;
    % If not possible, try to insert other relation
    insertRelation(SrcElem,IntermedElem,IntermedElemType,NewElem,NewElemType)
  ),
  (
    (
      % Verify fold success
      utilises(SrcElem,DstElem,_) 
    )
    ;
    ( 
      % Else try to attach new element to a existing element
      findall(LinkElem,linkElementToGraph(SrcElem,DstElem,NewElem,NewElemType,LinkElem),_),
      (insertUses(SrcElem,DstElem,NewElem,NewElemType,_,_);true)
    )
    ;
    (
      % Else create new element
      gensym('99',StrNewElem2),
      ato-to-int(StrNewElem2,NewElem2),
      % And continue fold with new element
      insertFold(SrcElem,SrcElemType,DstElem,DstElemType,NewElem,NewElemType,NewElem2) 
    )
  ),
  % Verify fold success
  utilises(SrcElem,DstElem,_).


%--------------------------------------------------------------------------
% Verify fold success
%
% utilises(+Src_element,+Dst_element,-Fold_element)
%--------------------------------------------------------------------------
utilises(SrcElem,DstElem,FoldElem) :-
  (                                                       % Uses
    convUses(SrcElem,FoldElem),                           % Is there a fold element ?
    convUses(FoldElem,DstElem),
    verifyConstrRelation(FoldElem,DstElem),               % That satisfy decoupling constraints ?
    convNode(PkgName,1),                                  
    pathExist(PkgName,FoldElem)                           % Contained in a access graph ?
  )
  ;
  (                                                       % Virtual Uses
    convUses(SrcElem,FoldElem),                           % Is there a fold element ?
    convNode(DstElem,DstElemType),
    (
      (DstElemType=2;DstElemType=3) *->                   % If fold element is class or interface
          inheritsLink(DstElem,FoldElem)                  % And inherits dst element
          ;                                               % Or
          convContains(ContFoldElem,FoldElem),            % Get class of fold element 
          convContains(ContDstElem,DstElem),              % Get class of dst element
          ContDstElem\=ContFoldElem,                        
          inheritsLink(ContDstElem,ContFoldElem)          % Dst element class inherit fold elem class ?
    )
  ).


%--------------------------------------------------------------------------
% Link element to a graph
%
% linkElementToGraph(+Src_element,+Dst_element,+New_element,
%  +New_element_type,-Link_element)
%--------------------------------------------------------------------------
linkElementToGraph(SrcElem,DstElem,NewElem,NewElemType,LinkElem) :-
  % Find all possible destination type
  (getAllPossibleGraphRelation(NewToLinkElemRel,LinkElemType,NewElemType),
    label([NewToLinkElemRel,LinkElemType])),
  convNode(LinkElem,LinkElemType),                        % Get nodes with possible dest type
  (
    (
    inClassSet(LinkElem,DstElem)                          % Test if link element is in class set
    )
    ;
    (
    pathExist(LinkElem,DstElem)                           % Path exist between nodes and final dst
    )
  ),
  (NewToLinkElemRel\=1 *->                                % If relation with link element is not inherits
    verifyConstrRelation(SrcElem,LinkElem);true),         % Verify that relation satisfy decoupling constr
  createRelation(NewToLinkElemRel,LinkElem,NewElem).      % Create relation if link element found
  

%--------------------------------------------------------------------------
% Insert uses
%
% insertUses(+Src_element,+Dst_element,+Intermediate_element,+Intermediate_element_type,
% +New_element,-New_element_type)
%--------------------------------------------------------------------------
insertUses(SrcElem,DstElem,IntermedElem,IntermedElemType,NewElem,NewElemType) :-
  (
    SrcElem==IntermedElem,                                % If first inserted element 
    methodSet(ClassSet,MethodSet,MethodId),               % Get method set
    MethodId\=0,
    convContains(ClassDstElem,DstElem),                   % Get dst elem class
    convContains(ClassMethodId,MethodId),                 % Get methodSet method class
    member(ClassDstElem,ClassSet),                        % If dst elem class is in class set
    member(DstElem,MethodSet),                            % And dst element is in method set
    (convUses(SrcElem,MethodId) *->                       % If uses relation exist
      true
      ;createRelation(3,SrcElem,MethodId)),               % Else create use relation
    (convInherits(ClassDstElem,ClassMethodId) *->         % If inherits relation exist
      true
      ;createRelation(1,ClassDstElem,ClassMethodId))      % Else create inherits relation
  )
  ;
  (
    SrcElem==IntermedElem,                                % If first inserted element
    convNode(DstElem,DstElemType),
    (DstElemType=2;DstElemType=3),                        % And dst elem type is class or interface
    convInherits(DstElem,InterfaceDstElem),               % And dst elem inherits interface
    convNode(InterfaceDstElem,InterfaceDstElemType),
    InterfaceDstElemType=3,
    createRelation(3,SrcElem,InterfaceDstElem)            % Create relation with interface
  )
  ;
  (SrcElem==IntermedElem *->                              % If first inserted element
    (getAllPossibleGraphRelation(3,IntermedElemType,NewElemType)
      ,label([NewElemType])),
    (NewElemType=3;NewElemType=5),                        % Set new element type (inteface or method)
    verifyConstrRelation(IntermedElem,NewElem),           % Verify that relation satisfy constraints
    createElement(NewElem,NewElemType),                   % Create element
    createRelation(3,IntermedElem,NewElem)                % Create uses relation
    ;
    convUses(SrcElem,IntermedElem),                       % Else
    verifyConstrRelation(IntermedElem,DstElem),           % Try static fold
    createRelation(3,IntermedElem,DstElem)                % If ok create uses relation
  ).


%--------------------------------------------------------------------------
% Insert Relation
%
%  
% insertRelation(+Src_element,+Intermediate_element,+Intermediate_element_type,
% +New_element,-New_element_type)
%--------------------------------------------------------------------------
insertRelation(SrcElem,IntermedElem,IntermedElemType,NewElem,NewElemType) :-
  SrcElem\=IntermedElem,                                  % If not first relation 
  % Get all possible relation
  (getAllPossibleGraphRelation(SrcToNewRel,NewElemType,IntermedElemType)
    ,label([SrcToNewRel,NewElemType])),
  SrcToNewRel\=3,                                         % Not uses relation 
  (NewElemType=3;NewElemType=5),                          % New elem type in class or method
  createElement(NewElem,NewElemType),                     % Create element
  createRelation(SrcToNewRel,NewElem,IntermedElem).       % Create relation

 
%--------------------------------------------------------------------------
% Create element
%
% createElement(+New_element,+New_element_type) 
%--------------------------------------------------------------------------
createElement(NewElem,NewElemType) :-
  retractall(convNode(NewElem,_)),                        % Remove inserted elem (backtrack)
  retractall(convInherits(_,NewElem)),                    % Remove inserted relation (backtrack)
  retractall(convContains(_,NewElem)),                    % Remove inserted relation (backtrack)
  retractall(convUses(_,NewElem)),                        % Remove inserted relation (backtrack) 
  assert(convNode(NewElem,NewElemType)).                  % Insert new node 

%--------------------------------------------------------------------------
% Create relation
%
% createRelation(+Relation_type,+Src_element,+Dst_element) 
%--------------------------------------------------------------------------
createRelation(RelType,SrcElem,DstElem) :-
  (RelType=1,\+convInherits(SrcElem,DstElem) *->
    assert(convInherits(SrcElem,DstElem));true),
  (RelType=2,\+convContains(SrcElem,DstElem) *->
    assert(convContains(SrcElem,DstElem));true),
  (RelType=3,\+convUses(SrcElem,DstElem) *->
    assert(convUses(SrcElem,DstElem));true).
  

